import { Logger } from '$lib/log';
import { TimelineReader } from '$lib/server/timeline';
import type { Handle, HandleServerError } from '@sveltejs/kit';
import { error } from '@sveltejs/kit';
import fs from 'fs/promises';
import { close } from '$lib/server/db';
import { version } from '$app/environment';

const logger = new Logger('App');

if (process?.pid) {
  try {
    await fs.writeFile('moshing-mammut.pid', process.pid.toString());
  } catch (e) {
    logger.error('Could not write PID to file', e);
  }
}

logger.log('App startup, version', version, 'PID', process?.pid);

logger.log('Debug log enabled', Logger.isDebugEnabled());
TimelineReader.init();

process.on('sveltekit:shutdown', (reason) => {
  close();
  logger.log('Shutting down', reason);
  process.exit(0);
});

export const handleError = (({ error, status }) => {
  if (error instanceof Error && status !== 404) {
    logger.error('Something went wrong:', error.name, error.message);
  }

  return {
    message: `Something went wrong! ${error}`
  };
}) satisfies HandleServerError;

export const handle = (async ({ event, resolve }) => {
  // Reeder *insists* on checking /feed instead of /feed.xml
  if (event.url.pathname === '/feed') {
    return new Response('', { status: 301, headers: { Location: '/feed.xml' } });
  }
  if (event.url.pathname === '/feed.xml') {
    const f = await fs.readFile('feed.xml', { encoding: 'utf8' });
    return new Response(f, { headers: [['Content-Type', 'application/atom+xml']] });
  }

  // Ideally, this would be served by apache
  if (event.url.pathname.startsWith('/avatars/')) {
    const fileName = event.url.pathname.split('/').pop() ?? 'unknown.jpeg';
    const suffix = fileName.split('.').pop() ?? 'jpeg';
    try {
      //This should work, but doesn't yet. See: https://github.com/nodejs/node/issues/45853
      /*
      const stat = await fs.stat('avatars/' + fileName);
      const fd = await fs.open('avatars/' + fileName);
      const readStream = fd
        .readableWebStream()
        .getReader({ mode: 'byob' }) as ReadableStream<Uint8Array>;
      logger.info('sending. size: ', stat.size);
      return new Response(readStream, {
        headers: [
          ['Content-Type', 'image/' + suffix],
          ['Content-Length', stat.size.toString()]
        ]
      });
      */

      const f = await fs.readFile('avatars/' + fileName);
      return new Response(f, { headers: [['Content-Type', 'image/' + suffix]] });
    } catch (e) {
      logger.error('no stream', e);
      error(404);
    }
  }

  const response = await resolve(event);
  return response;
}) satisfies Handle;
