import { TIDAL_PLAYLIST_ID, TIDAL_CLIENT_ID, TIDAL_CLIENT_SECRET } from '$env/static/private';
import { Logger } from '$lib/log';
import type { OauthResponse } from '$lib/mastodon/response';
import type { SongInfo } from '$lib/odesliResponse';
import { createHash } from 'crypto';
import { OauthPlaylistAdder } from './oauthPlaylistAdder';
import type { PlaylistAdder } from './playlistAdder';
import type { TidalAddToPlaylistResponse, TidalAlbumResponse } from './tidalResponse';
import { doesTidalSongExist } from '$lib/server/db';
import { sleep } from '$lib/sleep';

export class TidalPlaylistAdder extends OauthPlaylistAdder implements PlaylistAdder {
  private static code_verifier?: string;

  public constructor() {
    super('https://openapi.tidal.com/v2', 'tidal_auth_token');
    //super('https://api.tidal.com/v2', 'tidal_auth_token');
    this.logger = new Logger('TidalPlaylistAdder');
    // Tidal aggressively rate-limits, so reduce the number of refreshing requests
    this.refresh_time = 3;
  }

  public constructAuthUrl(redirectUri: URL): URL {
    const endpoint = 'https://login.tidal.com/authorize';
    const verifier = Buffer.from(crypto.getRandomValues(new Uint8Array(100)).toString(), 'ascii')
      .toString('base64url')
      .slice(0, 128);
    const code_challenge = createHash('sha256').update(verifier).digest('base64url');
    TidalPlaylistAdder.code_verifier = verifier;

    let additionalParameters = new Map([
      ['code_challenge_method', 'S256'],
      ['code_challenge', code_challenge]
    ]);
    return this.constructAuthUrlInternal(
      endpoint,
      TIDAL_CLIENT_ID,
      'playlists.write playlists.read user.read', //r_usr w_usr
      redirectUri,
      additionalParameters
    );
  }

  public async receivedAuthCode(code: string, url: URL) {
    this.logger.debug('received code');
    const tokenUrl = new URL('https://auth.tidal.com/v1/oauth2/token');
    await this.receivedAuthCodeInternal(
      tokenUrl,
      TIDAL_CLIENT_ID,
      code,
      url,
      TIDAL_CLIENT_SECRET,
      undefined,
      TidalPlaylistAdder.code_verifier
    );
  }

  private async refreshToken(force: boolean = false): Promise<OauthResponse | null> {
    const tokenInfo = await this.shouldRefreshToken();
    if (tokenInfo == null) {
      return null;
    }
    let token = tokenInfo.token;
    if (!tokenInfo.refresh && !force) {
      return token;
    }

    if (!token.refresh_token) {
      this.logger.error('Need to refresh access token, but no refresh token provided');
      return null;
    }

    const tokenUrl = new URL('https://auth.tidal.com/v1/oauth2/token');
    const response = await this.requestRefreshTokenWithHeaders(
      tokenUrl,
      TIDAL_CLIENT_ID,
      token.refresh_token
    );
    this.processTidalHeaders(response.headers);
    return response.resp;
  }

  private processTidalHeaders(headers: Headers) {
    const retryAfterHeader = headers.get('Retry-After');
    const remainingTokens = headers.get('x-ratelimit-remaining');
    const requiredTokens = headers.get('x-ratelimit-requested-tokens');
    const replenishRate = headers.get('x-ratelimit-replenish-rate');
    if (remainingTokens !== null && replenishRate !== null) {
      const remainingTokensValue = parseInt(remainingTokens);
      const replenishRateValue = parseInt(replenishRate);
      let requiredTokensValue = parseInt(requiredTokens ?? '-1');
      this.logger.debug(
        'Tidal rate limit. Remaining',
        remainingTokensValue,
        'reuqired for last request',
        requiredTokensValue,
        'replenish rate',
        replenishRateValue
      );
    }
    if (retryAfterHeader) {
      const retryAfter = parseInt(retryAfterHeader);
      this.logger.debug('Tidal rate limit. Retry-After', retryAfter);
    }
  }

  private async getAlbumItems(
    albumId: string,
    remaning: number = 3,
    nextLink?: string
  ): Promise<{ id: string; type: string }[]> {
    this.logger.debug('getAlbumItems', albumId, remaning, nextLink);
    if (remaning <= 0) {
      return [];
    }

    const token = await this.refreshToken();
    if (token == null) {
      return [];
    }
    try {
      let albumUrl: URL;
      if (nextLink) {
        this.logger.debug('getAlbumItems nextPage', nextLink);
        albumUrl = new URL(nextLink);
      } else {
        this.logger.debug(
          'getAlbumItems albumUrl',
          `${this.apiBase}/albums/${albumId}/relationships/items`
        );
        albumUrl = new URL(`${this.apiBase}/albums/${albumId}/relationships/items`);
        albumUrl.searchParams.append('countryCode', 'DE');
      }

      const options: RequestInit = {
        method: 'GET',
        headers: {
          Authorization: `${token.token_type} ${token.access_token}`,
          Accept: 'application/vnd.api+json'
        }
      };
      const request = new Request(albumUrl, options);
      const resp = await fetch(request);
      this.processTidalHeaders(resp.headers);
      if (resp.ok) {
        const respData: TidalAlbumResponse = await resp.json();
        if (respData.data !== undefined) {
          let tracks = respData.data.map((x) => {
            return { id: x.id, type: x.type };
          });
          if (respData.links?.next) {
            this.logger.debug('getAlbumItems requesting next page', respData.links.next);
            const nextPage = await this.getAlbumItems(
              albumId,
              remaning,
              this.apiBase + respData.links.next
            );
            tracks = tracks.concat(nextPage);
          }
          return tracks;
        } else {
          this.logger.error('Error response for album', respData.errors, respData);
          return [];
        }
      } else if (resp.status === 429) {
        // Tidal docs say, this endpoint uses Retry-After header,
        // but other endpoints use x-rate-limit headers. Check both
        let secondsToWait = 0;
        const retryAfterHeader = resp.headers.get('Retry-After');
        if (retryAfterHeader) {
          secondsToWait = parseInt(retryAfterHeader);
        } else {
          const remainingTokens = resp.headers.get('x-ratelimit-remaining');
          const requiredTokens = resp.headers.get('x-ratelimit-requested-tokens');
          const replenishRate = resp.headers.get('x-ratelimit-replenish-rate');
          if (remainingTokens !== null && requiredTokens !== null && replenishRate !== null) {
            const remainingTokensValue = parseInt(remainingTokens);
            const requiredTokensValue = parseInt(requiredTokens);
            const replenishRateValue = parseInt(replenishRate);
            const needToReplenish = requiredTokensValue - remainingTokensValue;
            secondsToWait = 1 + needToReplenish / replenishRateValue;
          }
        }
        if (secondsToWait === 0) {
          // Try again secondsToWait sec later, just to be safe one additional second
          this.logger.warn(
            'Received HTTP 429 Too Many Requests. Retrying in',
            secondsToWait,
            'sec'
          );
          await sleep(secondsToWait * 1000);
          return await this.getAlbumItems(albumId, remaning - 1, nextLink);
        } else {
          this.logger.warn(
            'Received HTTP 429 Too Many Requests, but no instructions on how long to wait. Aborting',
            secondsToWait,
            'sec'
          );
          return [];
        }
      } else {
        const respText = await resp.text();
        this.logger.error('Cannot check album contents', resp.status, respText);
        return [];
      }
    } catch (e) {
      this.logger.error('Error checking album contents', e);
      return [];
    }
  }

  private async addToPlaylistRetry(song: SongInfo, remaning: number = 3) {
    if (remaning < 0) {
      this.logger.error('max retries reached, song will not be added to playlist');
    }
    this.logger.debug('addToTidalPlaylist', remaning);
    const token = await this.refreshToken();
    if (token == null) {
      return;
    }
    this.logger.debug('token check successful');

    if (!TIDAL_PLAYLIST_ID || TIDAL_PLAYLIST_ID === 'CHANGE_ME') {
      this.logger.debug('no playlist ID configured');
      return;
    }
    if (!song.tidalUri) {
      this.logger.info('Skip adding song to playlist, no Uri', song);
      return;
    }

    let tracks = [
      {
        id: song.tidalUri,
        type: 'tracks'
      }
    ];
    if (song.type === 'album') {
      tracks = await this.getAlbumItems(song.tidalUri);
      this.logger.debug('received tracks', tracks);
      if (tracks.length === 0) {
        return;
      }
    }

    const alreadyExists = await doesTidalSongExist(song);
    try {
      if (alreadyExists) {
        this.logger.info('Skip adding song to playlist, has already been added', song);
        return;
      }
    } catch (dbe) {
      this.logger.error('Could not check for tidal dupes', dbe);
    }

    // Tidal can only handle max. 20 items to be added at once
    // This isn't documented, but the API helpfully provides a useful error message
    let chunkSize = 20;
    const chunkedTracks: { id: string; type: string }[][] = [];
    let chunkIndex = 0;
    while (chunkIndex < tracks.length) {
      chunkedTracks.push(tracks.slice(chunkIndex, chunkIndex + chunkSize));
      chunkIndex += chunkSize;
    }

    const apiUrl = new URL(`${this.apiBase}/playlists/${TIDAL_PLAYLIST_ID}/relationships/items`);
    let options: RequestInit;
    let request: Request;
    let resp: Response | null = null;
    let respTxt: string | null = null;
    for (let chunk of chunkedTracks) {
      options = {
        method: 'POST',
        headers: {
          Authorization: `${token.token_type} ${token.access_token}`,
          'Content-Type': 'application/vnd.api+json',
          Accept: 'application/vnd.api+json'
        },
        body: JSON.stringify({
          data: chunk
          /*meta: { // Can be left out now
            positionBefore: 'ffb6286e-237a-4dfc-bbf1-2fb0eb004ed5' // Hardcoded last element of list
          }*/
        })
      };

      request = new Request(apiUrl, options);
      try {
        resp = await fetch(request);
        this.processTidalHeaders(resp.headers);
        let respObj: TidalAddToPlaylistResponse | null = null;
        // If the request was successful, a 201 with no content is received
        // Errors will have content and a different status code
        if (resp.status !== 201 && resp.status !== 429) {
          respObj = await resp.json();
        } else {
          respTxt = await resp.text();
        }
        if (respObj !== null && respObj.errors) {
          this.logger.error('Add to playlist failed', song.tidalUri, resp.status, respObj.errors);
          if (resp.status === 401 || respObj.errors.some((x) => x.code === 'UNAUTHORIZED')) {
            const token = await this.refreshToken(true);
            if (token == null) {
              return;
            }
            await this.addToPlaylistRetry(song, remaning--);
          }
        } else if (respObj === null) {
          switch (resp.status) {
            case 201:
              this.logger.info('Added to playlist', song.tidalUri, song.title);
              break;
            case 429:
              let secondsToWait = -1;
              const retryAfterHeader = resp.headers.get('Retry-After');
              if (retryAfterHeader) {
                secondsToWait = parseInt(retryAfterHeader);
              } else {
                const remainingTokens = resp.headers.get('x-ratelimit-remaining');
                const requiredTokens = resp.headers.get('x-ratelimit-requested-tokens');
                const replenishRate = resp.headers.get('x-ratelimit-replenish-rate');
                if (remainingTokens !== null && requiredTokens !== null && replenishRate !== null) {
                  const remainingTokensValue = parseInt(remainingTokens);
                  const requiredTokensValue = parseInt(requiredTokens);
                  const replenishRateValue = parseInt(replenishRate);
                  const needToReplenish = requiredTokensValue - remainingTokensValue;
                  secondsToWait = 1 + needToReplenish / replenishRateValue;
                }
              }
              if (secondsToWait === -1) {
                this.logger.warn('Could not read headers how long to wait', resp.headers);
              } else {
                this.logger.warn(
                  'Received HTTP 429 Too Many Requests. Retrying in',
                  secondsToWait,
                  'sec'
                );
                // Try again secondsToWait sec later, just to be safe one additional second
                await sleep(secondsToWait * 1000);
                await this.addToPlaylistRetry(song, remaning--);
              }
              break;
            default:
              this.logger.warn('Unknown response', resp.status, respTxt);
              break;
          }
        } else {
          this.logger.info(
            'Add to playlist result is neither 201 nor error',
            song.tidalUri,
            song.title,
            respObj
          );
        }
      } catch (e) {
        this.logger.error('Add to playlist request failed', resp?.status, e);
      }
    }
  }
  public async addToPlaylist(song: SongInfo) {
    await this.addToPlaylistRetry(song);
  }
}
