import { YOUTUBE_CLIENT_ID, YOUTUBE_CLIENT_SECRET, YOUTUBE_PLAYLIST_ID } from '$env/static/private';
import { Logger } from '$lib/log';
import type { OauthResponse } from '$lib/mastodon/response';
import type {
  YoutubePlaylistItemResponse,
  YoutubeResponse
} from '$lib/server/playlist/youtubeResponse';
import type { SongInfo } from '$lib/odesliResponse';
import { OauthPlaylistAdder } from './oauthPlaylistAdder';
import type { PlaylistAdder } from './playlistAdder';

export class YoutubePlaylistAdder extends OauthPlaylistAdder implements PlaylistAdder {
  public constructor() {
    super('https://www.googleapis.com/youtube/v3', 'yt_auth_token');
    this.logger = new Logger('YoutubePlaylistAdder');
  }

  public constructAuthUrl(redirectUri: URL): URL {
    this.redirectUri = redirectUri;
    let additionalParameters = new Map([
      ['access_type', 'offline'],
      ['include_granted_scopes', 'false']
    ]);
    const endpoint = 'https://accounts.google.com/o/oauth2/v2/auth';
    return this.constructAuthUrlInternal(
      endpoint,
      YOUTUBE_CLIENT_ID,
      'https://www.googleapis.com/auth/youtube',
      redirectUri,
      additionalParameters
    );
  }

  public async receivedAuthCode(code: string, url: URL) {
    this.logger.debug('received code');
    this.redirectUri = url;
    const tokenUrl = new URL('https://oauth2.googleapis.com/token');
    await this.receivedAuthCodeInternal(
      tokenUrl,
      YOUTUBE_CLIENT_ID,
      code,
      url,
      YOUTUBE_CLIENT_SECRET
    );
    const token = await this.refreshToken();
    if (token == null) {
      return;
    }
  }

  private async checkAuthorizedToken(token: OauthResponse) {
    try {
      this.logger.debug('Checking authorized token');
      const res = await fetch(this.apiBase + '/channels?part=id&mine=true', {
        method: 'GET',
        headers: { Authorization: `${token.token_type} ${token.access_token}` }
      }).then((r) => r.json());
      this.logger.debug('Checked authorized token', res);
    } catch (e) {
      this.logger.error('Error checking authorized token', e);
    }
  }

  private async refreshToken(force: boolean = false): Promise<OauthResponse | null> {
    const tokenInfo = await this.shouldRefreshToken();
    if (tokenInfo == null) {
      return null;
    }
    let token = tokenInfo.token;
    if (!tokenInfo.refresh && !force) {
      return token;
    }
    if (!token.refresh_token) {
      this.logger.error('Need to refresh access token, but no refresh token provided');
      return null;
    }

    const tokenUrl = new URL('https://oauth2.googleapis.com/token');
    const refreshedToken = await this.requestRefreshToken(
      tokenUrl,
      YOUTUBE_CLIENT_ID,
      token.refresh_token,
      this.getRedirectUri('ytauth').toString(),
      YOUTUBE_CLIENT_SECRET
    );
    if (refreshedToken !== null) {
      await this.checkAuthorizedToken(refreshedToken);
    }
    return refreshedToken;
  }

  private async getVideoUris(song: SongInfo, token: OauthResponse): Promise<string[]> {
    this.logger.debug('getVideoUris');

    if (!song.youtubeUrl) {
      this.logger.info('Skip adding song to YT playlist, no youtube Url', song);
      return [];
    }

    const songUrl = new URL(song.youtubeUrl);
    let videoIds: string[] = [];
    if (song.type === 'song') {
      const youtubeId = songUrl.searchParams.get('v');
      if (!youtubeId) {
        this.logger.debug(
          'Skip adding song to YT playlist, could not extract YT id from URL',
          song.youtubeUrl
        );
        return [];
      }
      this.logger.debug('Found YT id from URL', song.youtubeUrl, youtubeId);
      return [youtubeId];
    }

    const albumPlaylistId = songUrl.searchParams.get('list') ?? '';
    const albumItemsUrl = new URL(this.apiBase + '/playlistItems');
    albumItemsUrl.searchParams.append('maxResults', '50');
    albumItemsUrl.searchParams.append('playlistId', albumPlaylistId);
    albumItemsUrl.searchParams.append('part', 'snippet');
    const albumPlaylistItemRequest = new Request(albumItemsUrl, {
      headers: { Authorization: `${token.token_type} ${token.access_token}` }
    });

    const albumPlaylistItem = await this.fetchWithRetry<YoutubePlaylistItemResponse>(
      albumPlaylistItemRequest,
      token
    );
    if (albumPlaylistItem === null) {
      this.logger.info('Could not check album tracks');
      return [];
    }
    const albumTracks = albumPlaylistItem.items ?? [];
    videoIds = albumTracks.map((x) => x.snippet?.resourceId?.videoId).filter((x) => x);
    this.logger.info(
      'Found',
      albumPlaylistItem.pageInfo?.totalResults,
      'songs in album, received',
      albumTracks.length
    );
    if (videoIds.length === 0) {
      this.logger.debug(
        'Skip adding album to YT playlist, is empty',
        song.youtubeUrl,
        albumPlaylistItem
      );
    }
    return videoIds;
  }
  public async addToPlaylist(song: SongInfo) {
    this.logger.debug('addToYoutubePlaylist');
    let token = await this.refreshToken();
    if (token == null) {
      return;
    }

    if (!YOUTUBE_PLAYLIST_ID || YOUTUBE_PLAYLIST_ID === 'CHANGE_ME') {
      this.logger.debug('no playlist ID configured');
      return;
    }

    const videoIds = await this.getVideoUris(song, token);

    for (let youtubeId of videoIds) {
      this.logger.debug('Adding to playlist', youtubeId);
      const alreadyAdded = await this.isVideoInPlaylist(youtubeId, token);
      if (alreadyAdded) {
        this.logger.info('Item already in playlist', song.youtubeUrl, song.title);
        continue;
      } else {
        this.logger.debug('Item not already in playlist', song.youtubeUrl, song.title);
      }

      const success = await this.addSongToPlaylist(youtubeId, token);
      if (success) {
        this.logger.info('Added to playlist', youtubeId, song.title);
      }
    }
  }

  private async addSongToPlaylist(videoId: string, token: OauthResponse): Promise<boolean> {
    const addItemUrl = new URL(this.apiBase + '/playlistItems');
    addItemUrl.searchParams.append('part', 'snippet');
    const options: RequestInit = {
      method: 'POST',
      headers: { Authorization: `${token.token_type} ${token.access_token}` },
      body: JSON.stringify({
        snippet: {
          playlistId: YOUTUBE_PLAYLIST_ID,
          resourceId: {
            videoId: videoId,
            kind: 'youtube#video'
          }
        }
      })
    };
    const request = new Request(addItemUrl, options);
    const resp = await this.fetchWithRetry(request, token);
    return resp !== null;
  }

  private async fetchWithRetry<T extends YoutubeResponse>(
    request: Request,
    token: OauthResponse,
    retries: number = 3
  ): Promise<T | null> {
    if (retries <= 0) {
      return null;
    }
    request.headers.set('Authorization', `${token.token_type} ${token.access_token}`);
    const resp = await fetch(request);
    let respObj: T = await resp.json();
    if (!respObj.error) {
      return respObj;
    }
    this.logger.error(
      'Request failed',
      respObj.error.errors,
      respObj.error.code,
      respObj.error.message
    );
    if (
      respObj.error.code === 401 ||
      respObj.error.code === 403 ||
      respObj.error.errors.some((x) => x.reason === 'authError')
    ) {
      this.logger.info('Refreshing auth token');
      const newToken = await this.refreshToken(true);
      if (newToken == null) {
        this.logger.error('Refreshing auth token failed');
        return null;
      }
      return await this.fetchWithRetry(request, newToken, retries - 1);
    }
    return null;
  }

  private async isVideoInPlaylist(videoId: string, token: OauthResponse): Promise<boolean> {
    const playlistItemsUrl = new URL(this.apiBase + '/playlistItems');
    playlistItemsUrl.searchParams.append('videoId', videoId);
    playlistItemsUrl.searchParams.append('playlistId', YOUTUBE_PLAYLIST_ID);
    playlistItemsUrl.searchParams.append('part', 'id');
    let existingPlaylistItemRequest = new Request(playlistItemsUrl, {
      headers: { Authorization: `${token.token_type} ${token.access_token}` }
    });

    const existingPlaylistItem = await this.fetchWithRetry<YoutubePlaylistItemResponse>(
      existingPlaylistItemRequest,
      token
    );
    if (existingPlaylistItem === null) {
      this.logger.error('Could not check if item is already in playlist');
      return false;
    }
    return existingPlaylistItem.pageInfo && existingPlaylistItem.pageInfo.totalResults > 0;
  }
}
